@@ -1,16 +1,25 @@
 import os
 import tempfile
 import time
-from typing import Dict, List, Union
+import warnings
+from typing import Dict, List, Optional, Union
 
-from blockfrost import ApiUrls, BlockFrostApi
+from blockfrost import ApiError, ApiUrls, BlockFrostApi
+from blockfrost.utils import Namespace
 
 from pycardano.address import Address
-from pycardano.backend.base import ChainContext, GenesisParameters, ProtocolParameters
+from pycardano.backend.base import (
+    ALONZO_COINS_PER_UTXO_WORD,
+    ChainContext,
+    GenesisParameters,
+    ProtocolParameters,
+)
 from pycardano.exception import TransactionFailedException
 from pycardano.hash import SCRIPT_HASH_SIZE, DatumHash, ScriptHash
+from pycardano.nativescript import NativeScript
 from pycardano.network import Network
-from pycardano.plutus import ExecutionUnits
+from pycardano.plutus import ExecutionUnits, PlutusV1Script, PlutusV2Script
+from pycardano.serialization import RawCBOR
 from pycardano.transaction import (
     Asset,
     AssetName,
@@ -20,6 +29,7 @@ from pycardano.transaction import (
     UTxO,
     Value,
 )
+from pycardano.types import JsonDict
 
 __all__ = ["BlockFrostChainContext"]
 
@@ -30,13 +40,34 @@ class BlockFrostChainContext(ChainContext):
     Args:
         project_id (str): A BlockFrost project ID obtained from https://blockfrost.io.
         network (Network): Network to use.
+        base_url (str): Base URL for the BlockFrost API. Defaults to the preprod url.
     """
 
-    def __init__(self, project_id: str, network: Network = Network.TESTNET):
-        self._network = network
+    api: BlockFrostApi
+    _epoch_info: Namespace
+    _epoch: Optional[int] = None
+    _genesis_param: Optional[GenesisParameters] = None
+    _protocol_param: Optional[ProtocolParameters] = None
+
+    def __init__(
+        self,
+        project_id: str,
+        network: Optional[Network] = None,
+        base_url: str = ApiUrls.preprod.value,
+    ):
+        if network is not None:
+            warnings.warn(
+                "`network` argument will be deprecated in the future. Directly passing `base_url` is recommended."
+            )
+            self._network = network
+        else:
+            self._network = Network.TESTNET
+
         self._project_id = project_id
         self._base_url = (
-            ApiUrls.testnet.value
+            base_url
+            if base_url
+            else ApiUrls.preprod.value
             if self.network == Network.TESTNET
             else ApiUrls.mainnet.value
         )
@@ -60,7 +91,8 @@ class BlockFrostChainContext(ChainContext):
     @property
     def epoch(self) -> int:
         if not self._epoch or self._check_epoch_and_update():
-            self._epoch = self.api.epoch_latest().epoch
+            new_epoch: int = self.api.epoch_latest().epoch
+            self._epoch = new_epoch
         return self._epoch
 
     @property
@@ -95,6 +127,7 @@ class BlockFrostChainContext(ChainContext):
                 protocol_major_version=int(params.protocol_major_ver),
                 protocol_minor_version=int(params.protocol_minor_ver),
                 min_utxo=int(params.min_utxo),
+                min_pool_cost=int(params.min_pool_cost),
                 price_mem=float(params.price_mem),
                 price_step=float(params.price_step),
                 max_tx_ex_mem=int(params.max_tx_ex_mem),
@@ -104,10 +137,29 @@ class BlockFrostChainContext(ChainContext):
                 max_val_size=int(params.max_val_size),
                 collateral_percent=int(params.collateral_percent),
                 max_collateral_inputs=int(params.max_collateral_inputs),
-                coins_per_utxo_word=int(params.coins_per_utxo_word),
+                coins_per_utxo_word=int(params.coins_per_utxo_word)
+                or ALONZO_COINS_PER_UTXO_WORD,
+                coins_per_utxo_byte=int(params.coins_per_utxo_size),
+                cost_models={
+                    k: v.to_dict() for k, v in params.cost_models.to_dict().items()
+                },
             )
         return self._protocol_param
 
+    def _get_script(
+        self, script_hash: str
+    ) -> Union[PlutusV1Script, PlutusV2Script, NativeScript]:
+        script_type = self.api.script(script_hash).type
+        if script_type == "plutusV1":
+            return PlutusV1Script(bytes.fromhex(self.api.script_cbor(script_hash).cbor))
+        elif script_type == "plutusV2":
+            return PlutusV2Script(bytes.fromhex(self.api.script_cbor(script_hash).cbor))
+        else:
+            script_json: JsonDict = self.api.script_json(
+                script_hash, return_type="json"
+            )["json"]
+            return NativeScript.from_dict(script_json)
+
     def utxos(self, address: str) -> List[UTxO]:
         results = self.api.address_utxos(address, gather_pages=True)
 
@@ -118,7 +170,7 @@ class BlockFrostChainContext(ChainContext):
                 [result.tx_hash, result.output_index]
             )
             amount = result.amount
-            lovelace_amount = None
+            lovelace_amount = 0
             multi_assets = MultiAsset()
             for item in amount:
                 if item.unit == "lovelace":
@@ -133,33 +185,63 @@ class BlockFrostChainContext(ChainContext):
                         multi_assets[policy_id] = Asset()
                     multi_assets[policy_id][asset_name] = int(item.quantity)
 
+            amount = Value(lovelace_amount, multi_assets)
+
             datum_hash = (
-                DatumHash.from_primitive(result.data_hash) if result.data_hash else None
+                DatumHash.from_primitive(result.data_hash)
+                if result.data_hash and result.inline_datum is None
+                else None
             )
 
-            if not multi_assets:
-                tx_out = TransactionOutput(
-                    Address.from_primitive(address),
-                    amount=lovelace_amount,
-                    datum_hash=datum_hash,
-                )
-            else:
-                tx_out = TransactionOutput(
-                    Address.from_primitive(address),
-                    amount=Value(lovelace_amount, multi_assets),
-                    datum_hash=datum_hash,
-                )
+            datum = None
+
+            if hasattr(result, "inline_datum") and result.inline_datum is not None:
+                datum = RawCBOR(bytes.fromhex(result.inline_datum))
+
+            script = None
+
+            if (
+                hasattr(result, "reference_script_hash")
+                and result.reference_script_hash
+            ):
+                script = self._get_script(result.reference_script_hash)
+
+            tx_out = TransactionOutput(
+                Address.from_primitive(address),
+                amount=amount,
+                datum_hash=datum_hash,
+                datum=datum,
+                script=script,
+            )
             utxos.append(UTxO(tx_in, tx_out))
 
         return utxos
 
-    def submit_tx(self, cbor: Union[bytes, str]):
+    def submit_tx(self, cbor: Union[bytes, str]) -> str:
+        """Submit a transaction.
+
+        Args:
+            cbor (Union[bytes, str]): The serialized transaction to be submitted.
+
+        Returns:
+            str: The transaction hash.
+
+        Raises:
+            :class:`TransactionFailedException`: When fails to submit the transaction.
+        """
         if isinstance(cbor, str):
             cbor = bytes.fromhex(cbor)
         with tempfile.NamedTemporaryFile(delete=False) as f:
             f.write(cbor)
-        self.api.transaction_submit(f.name)
+        try:
+            response = self.api.transaction_submit(f.name)
+        except ApiError as e:
+            os.remove(f.name)
+            raise TransactionFailedException(
+                f"Failed to submit transaction. Error code: {e.status_code}. Error message: {e.message}"
+            ) from e
         os.remove(f.name)
+        return response
 
     def evaluate_tx(self, cbor: Union[bytes, str]) -> Dict[str, ExecutionUnits]:
         """Evaluate execution units of a transaction.
