@@ -1,8 +1,8 @@
 """All type of hashes in Cardano ledger spec."""
 
-from typing import TypeVar, Union
+from typing import Type, TypeVar, Union
 
-from pycardano.serialization import CBORSerializable
+from pycardano.serialization import CBORSerializable, limit_primitive_type
 
 __all__ = [
     "VERIFICATION_KEY_HASH_SIZE",
@@ -10,6 +10,7 @@ __all__ = [
     "TRANSACTION_HASH_SIZE",
     "DATUM_HASH_SIZE",
     "AUXILIARY_DATA_HASH_SIZE",
+    "POOL_KEY_HASH_SIZE",
     "SCRIPT_DATA_HASH_SIZE",
     "ConstrainedBytes",
     "VerificationKeyHash",
@@ -66,7 +67,8 @@ class ConstrainedBytes(CBORSerializable):
         return self.payload
 
     @classmethod
-    def from_primitive(cls: T, value: Union[bytes, str]) -> T:
+    @limit_primitive_type(bytes, str)
+    def from_primitive(cls: Type[T], value: Union[bytes, str]) -> T:
         if isinstance(value, str):
             value = bytes.fromhex(value)
         return cls(value)
