@@ -3,16 +3,17 @@
 from __future__ import annotations
 
 import re
-from collections import OrderedDict, defaultdict
+from collections import OrderedDict, UserList, defaultdict
 from copy import deepcopy
 from dataclasses import Field, dataclass, fields
 from datetime import datetime
 from decimal import Decimal
+from functools import wraps
 from inspect import isclass
-from pprint import pformat
-from typing import Any, Callable, ClassVar, List, Type, TypeVar, Union, get_type_hints
+from typing import Any, Callable, List, Optional, Type, TypeVar, Union, get_type_hints
 
 from cbor2 import CBOREncoder, CBORSimpleValue, CBORTag, dumps, loads, undefined
+from pprintpp import pformat
 from typeguard import check_type, typechecked
 
 from pycardano.exception import (
@@ -30,23 +31,53 @@ __all__ = [
     "ArrayCBORSerializable",
     "MapCBORSerializable",
     "DictCBORSerializable",
+    "RawCBOR",
     "list_hook",
+    "limit_primitive_type",
 ]
 
 
-class IndefiniteList:
-    def __init__(self, items):
-        self.items = items
-
-    def __eq__(self, other):
-        if isinstance(other, IndefiniteList):
-            return self.items == other.items
-        else:
-            return False
+def _identity(x):
+    return x
 
 
-Primitive = TypeVar(
-    "Primitive",
+class IndefiniteList(UserList):
+    def __init__(self, li: Primitive):  # type: ignore
+        super().__init__(li)  # type: ignore
+
+
+@dataclass
+class RawCBOR:
+    """A wrapper class for bytes that represents a CBOR value."""
+
+    cbor: bytes
+
+
+Primitive = Union[
+    bytes,
+    bytearray,
+    str,
+    int,
+    float,
+    Decimal,
+    bool,
+    None,
+    tuple,
+    list,
+    IndefiniteList,
+    dict,
+    defaultdict,
+    OrderedDict,
+    undefined.__class__,
+    datetime,
+    re.Pattern,
+    CBORSimpleValue,
+    CBORTag,
+    set,
+    frozenset,
+]
+
+PRIMITIVE_TYPES = (
     bytes,
     bytearray,
     str,
@@ -74,6 +105,31 @@ A list of types that could be encoded by
 `Cbor2 encoder <https://cbor2.readthedocs.io/en/latest/modules/encoder.html>`_ directly.
 """
 
+
+def limit_primitive_type(*allowed_types):
+    """
+    A helper function to validate primitive type given to from_primitive class methods
+
+    Not exposed to public by intention.
+    """
+
+    def decorator(func):
+        @wraps(func)
+        def wrapper(cls, value: Primitive):
+            if not isinstance(value, allowed_types):
+                allowed_types_str = [
+                    allowed_type.__name__ for allowed_type in allowed_types
+                ]
+                raise DeserializeException(
+                    f"{allowed_types_str} typed value is required for deserialization. Got {type(value)}: {value}"
+                )
+            return func(cls, value)
+
+        return wrapper
+
+    return decorator
+
+
 CBORBase = TypeVar("CBORBase", bound="CBORSerializable")
 
 
@@ -81,7 +137,7 @@ def default_encoder(
     encoder: CBOREncoder, value: Union[CBORSerializable, IndefiniteList]
 ):
     """A fallback function that encodes CBORSerializable to CBOR"""
-    assert isinstance(value, (CBORSerializable, IndefiniteList)), (
+    assert isinstance(value, (CBORSerializable, IndefiniteList, RawCBOR)), (
         f"Type of input value is not CBORSerializable, " f"got {type(value)} instead."
     )
     if isinstance(value, IndefiniteList):
@@ -89,9 +145,11 @@ def default_encoder(
         # handling here to explicitly write header (b'\x9f'), each body item, and footer (b'\xff') to
         # the output bytestring.
         encoder.write(b"\x9f")
-        for item in value.items:
+        for item in value:
             encoder.encode(item)
         encoder.write(b"\xff")
+    elif isinstance(value, RawCBOR):
+        encoder.write(value.cbor)
     else:
         encoder.encode(value.to_validated_primitive())
 
@@ -181,7 +239,7 @@ class CBORSerializable:
             elif isinstance(value, list):
                 return [_helper(k) for k in value]
             elif isinstance(value, IndefiniteList):
-                return IndefiniteList([_helper(k) for k in value.items])
+                return IndefiniteList([_helper(k) for k in value])
             elif isinstance(value, CBORTag):
                 return CBORTag(value.tag, _helper(value.value))
             else:
@@ -212,7 +270,7 @@ class CBORSerializable:
         return self.to_primitive()
 
     @classmethod
-    def from_primitive(cls: CBORBase, value: Primitive) -> CBORBase:
+    def from_primitive(cls: Type[CBORBase], value: Any) -> CBORBase:
         """Turn a CBOR primitive to its original class type.
 
         Args:
@@ -340,7 +398,7 @@ class CBORSerializable:
         return cls.from_primitive(value)
 
     def __repr__(self):
-        return pformat(vars(self))
+        return pformat(vars(self), indent=2)
 
 
 def _restore_dataclass_field(
@@ -359,9 +417,43 @@ def _restore_dataclass_field(
         return f.metadata["object_hook"](v)
     elif isclass(f.type) and issubclass(f.type, CBORSerializable):
         return f.type.from_primitive(v)
+    elif hasattr(f.type, "__origin__") and (f.type.__origin__ is list):
+        t_args = f.type.__args__
+        if len(t_args) != 1:
+            raise DeserializeException(
+                f"List types need exactly one type argument, but got {t_args}"
+            )
+        t = t_args[0]
+        if not isinstance(v, list):
+            raise DeserializeException(f"Expected type list but got {type(v)}")
+        if isclass(t) and issubclass(t, CBORSerializable):
+            return IndefiniteList([t.from_primitive(w) for w in v])
+        else:
+            return IndefiniteList(v)
     elif isclass(f.type) and issubclass(f.type, IndefiniteList):
         return IndefiniteList(v)
-    elif hasattr(f.type, "__origin__") and f.type.__origin__ is Union:
+    elif hasattr(f.type, "__origin__") and (f.type.__origin__ is dict):
+        t_args = f.type.__args__
+        if len(t_args) != 2:
+            raise DeserializeException(
+                f"Dict types need exactly two type arguments, but got {t_args}"
+            )
+        key_t = t_args[0]
+        val_t = t_args[1]
+        if isclass(key_t) and issubclass(key_t, CBORSerializable):
+            key_converter = key_t.from_primitive
+        else:
+            key_converter = _identity
+        if isclass(val_t) and issubclass(val_t, CBORSerializable):
+            val_converter = val_t.from_primitive
+        else:
+            val_converter = _identity
+        if not isinstance(v, dict):
+            raise DeserializeException(f"Expected dict type but got {type(v)}")
+        return {key_converter(key): val_converter(val) for key, val in v.items()}
+    elif hasattr(f.type, "__origin__") and (
+        f.type.__origin__ is Union or f.type.__origin__ is Optional
+    ):
         t_args = f.type.__args__
         for t in t_args:
             if isclass(t) and issubclass(t, IndefiniteList):
@@ -371,8 +463,11 @@ def _restore_dataclass_field(
                     return t.from_primitive(v)
                 except DeserializeException:
                     pass
-            elif t in Primitive.__constraints__ and isinstance(v, t):
-                return v
+            else:
+                if not isclass(t) and hasattr(t, "__origin__"):
+                    t = t.__origin__
+                if t in PRIMITIVE_TYPES and isinstance(v, t):
+                    return v
         raise DeserializeException(
             f"Cannot deserialize object: \n{v}\n in any valid type from {t_args}."
         )
@@ -443,8 +538,6 @@ class ArrayCBORSerializable(CBORSerializable):
         Test2(c='c', test1=Test1(a='a', b=None))
     """
 
-    field_sorter: ClassVar[Callable[[List], List]] = lambda x: x
-
     def to_shallow_primitive(self) -> List[Primitive]:
         """
         Returns:
@@ -455,7 +548,7 @@ class ArrayCBORSerializable(CBORSerializable):
                 types.
         """
         primitives = []
-        for f in self.__class__.field_sorter(fields(self)):
+        for f in fields(self):
             val = getattr(self, f.name)
             if val is None and f.metadata.get("optional"):
                 continue
@@ -463,7 +556,8 @@ class ArrayCBORSerializable(CBORSerializable):
         return primitives
 
     @classmethod
-    def from_primitive(cls: ArrayBase, values: List[Primitive]) -> ArrayBase:
+    @limit_primitive_type(list)
+    def from_primitive(cls: Type[ArrayBase], values: list) -> ArrayBase:
         """Restore a primitive value to its original class type.
 
         Args:
@@ -477,10 +571,6 @@ class ArrayCBORSerializable(CBORSerializable):
             DeserializeException: When the object could not be restored from primitives.
         """
         all_fields = [f for f in fields(cls) if f.init]
-        if type(values) != list:
-            raise DeserializeException(
-                f"Expect input value to be a list, got a {type(values)} instead."
-            )
 
         restored_vals = []
         type_hints = get_type_hints(cls)
@@ -514,7 +604,7 @@ class MapCBORSerializable(CBORSerializable):
 
         Basic usage:
 
-        >>> from dataclasses import dataclass
+        >>> from dataclasses import dataclass, field
         >>> @dataclass
         ... class Test1(MapCBORSerializable):
         ...     a: str=""
@@ -522,7 +612,7 @@ class MapCBORSerializable(CBORSerializable):
         >>> @dataclass
         ... class Test2(MapCBORSerializable):
         ...     c: str=None
-        ...     test1: Test1=Test1()
+        ...     test1: Test1=field(default_factory=Test1)
         >>> t = Test2(test1=Test1(a="a"))
         >>> t
         Test2(c=None, test1=Test1(a='a', b=''))
@@ -546,7 +636,7 @@ class MapCBORSerializable(CBORSerializable):
         >>> @dataclass
         ... class Test2(MapCBORSerializable):
         ...     c: str=field(default=None, metadata={"key": "0", "optional": True})
-        ...     test1: Test1=field(default=Test1(), metadata={"key": "1"})
+        ...     test1: Test1=field(default_factory=Test1, metadata={"key": "1"})
         >>> t = Test2(test1=Test1(a="a"))
         >>> t
         Test2(c=None, test1=Test1(a='a', b=''))
@@ -575,7 +665,8 @@ class MapCBORSerializable(CBORSerializable):
         return primitives
 
     @classmethod
-    def from_primitive(cls: MapBase, values: Primitive) -> MapBase:
+    @limit_primitive_type(dict)
+    def from_primitive(cls: Type[MapBase], values: dict) -> MapBase:
         """Restore a primitive value to its original class type.
 
         Args:
@@ -589,10 +680,6 @@ class MapCBORSerializable(CBORSerializable):
             :class:`pycardano.exception.DeserializeException`: When the object could not be restored from primitives.
         """
         all_fields = {f.metadata.get("key", f.name): f for f in fields(cls) if f.init}
-        if type(values) != dict:
-            raise DeserializeException(
-                f"Expect input value to be a dict, got a {type(values)} instead."
-            )
 
         kwargs = {}
         type_hints = get_type_hints(cls)
@@ -650,7 +737,7 @@ class DictCBORSerializable(CBORSerializable):
     def __getattr__(self, item):
         return getattr(self.data, item)
 
-    def __setitem__(self, key: KEY_TYPE, value: VALUE_TYPE):
+    def __setitem__(self, key: Any, value: Any):
         check_type("key", key, self.KEY_TYPE)
         check_type("value", value, self.VALUE_TYPE)
         self.data[key] = value
@@ -694,7 +781,8 @@ class DictCBORSerializable(CBORSerializable):
         return dict(sorted(self.data.items(), key=lambda x: _get_sortable_val(x[0])))
 
     @classmethod
-    def from_primitive(cls: DictBase, value: dict) -> DictBase:
+    @limit_primitive_type(dict)
+    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
         """Restore a primitive value to its original class type.
 
         Args:
@@ -709,12 +797,12 @@ class DictCBORSerializable(CBORSerializable):
         """
         if not value:
             raise DeserializeException(f"Cannot accept empty value {value}.")
+
         restored = cls()
         for k, v in value.items():
             k = (
                 cls.KEY_TYPE.from_primitive(k)
-                if isclass(cls.VALUE_TYPE)
-                and issubclass(cls.KEY_TYPE, CBORSerializable)
+                if isclass(cls.KEY_TYPE) and issubclass(cls.KEY_TYPE, CBORSerializable)
                 else k
             )
             v = (
@@ -726,13 +814,13 @@ class DictCBORSerializable(CBORSerializable):
             restored[k] = v
         return restored
 
-    def copy(self) -> DictBase:
+    def copy(self) -> DictCBORSerializable:
         return self.__class__(self)
 
 
 @typechecked
 def list_hook(
-    cls: Type[CBORSerializable],
+    cls: Type[CBORBase],
 ) -> Callable[[List[Primitive]], List[CBORBase]]:
     """A factory that generates a Callable which turns a list of Primitive to a list of CBORSerializables.
 
