@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-from copy import copy, deepcopy
+from copy import deepcopy
 from dataclasses import dataclass, field, fields
 from typing import Dict, List, Optional, Set, Tuple, Union
 
@@ -31,12 +31,18 @@ from pycardano.logging import logger
 from pycardano.metadata import AuxiliaryData
 from pycardano.nativescript import NativeScript, ScriptAll, ScriptAny, ScriptPubkey
 from pycardano.plutus import (
+    PLUTUS_V1_COST_MODEL,
+    PLUTUS_V2_COST_MODEL,
+    CostModels,
     Datum,
     ExecutionUnits,
+    PlutusV1Script,
+    PlutusV2Script,
     Redeemer,
     RedeemerTag,
+    ScriptType,
     datum_hash,
-    plutus_script_hash,
+    script_hash,
 )
 from pycardano.transaction import (
     Asset,
@@ -44,12 +50,13 @@ from pycardano.transaction import (
     MultiAsset,
     Transaction,
     TransactionBody,
+    TransactionInput,
     TransactionOutput,
     UTxO,
     Value,
     Withdrawals,
 )
-from pycardano.utils import fee, max_tx_fee, min_lovelace, script_data_hash
+from pycardano.utils import fee, max_tx_fee, min_lovelace_post_alonzo, script_data_hash
 from pycardano.witness import TransactionWitnessSet, VerificationKeyWitness
 
 __all__ = ["TransactionBuilder"]
@@ -81,29 +88,35 @@ class TransactionBuilder:
     execution_step_buffer: float = 0.2
     """Additional amount of execution step (in ratio) that will be added on top of estimation"""
 
-    ttl: int = field(default=None)
+    ttl: Optional[int] = field(default=None)
 
-    validity_start: int = field(default=None)
+    validity_start: Optional[int] = field(default=None)
 
-    auxiliary_data: AuxiliaryData = field(default=None)
+    auxiliary_data: Optional[AuxiliaryData] = field(default=None)
 
-    native_scripts: List[NativeScript] = field(default=None)
+    native_scripts: Optional[List[NativeScript]] = field(default=None)
 
-    mint: MultiAsset = field(default=None)
+    mint: Optional[MultiAsset] = field(default=None)
 
-    required_signers: List[VerificationKeyHash] = field(default=None)
+    required_signers: Optional[List[VerificationKeyHash]] = field(default=None)
 
     collaterals: List[UTxO] = field(default_factory=lambda: [])
 
-    certificates: List[Certificate] = field(default=None)
+    certificates: Optional[List[Certificate]] = field(default=None)
 
-    withdrawals: Withdrawals = field(default=None)
+    withdrawals: Optional[Withdrawals] = field(default=None)
+
+    reference_inputs: Set[Union[UTxO, TransactionInput]] = field(
+        init=False, default_factory=lambda: set()
+    )
 
     _inputs: List[UTxO] = field(init=False, default_factory=lambda: [])
 
     _excluded_inputs: List[UTxO] = field(init=False, default_factory=lambda: [])
 
-    _input_addresses: List[Address] = field(init=False, default_factory=lambda: [])
+    _input_addresses: List[Union[Address, str]] = field(
+        init=False, default_factory=lambda: []
+    )
 
     _outputs: List[TransactionOutput] = field(init=False, default_factory=lambda: [])
 
@@ -111,19 +124,27 @@ class TransactionBuilder:
 
     _datums: Dict[DatumHash, Datum] = field(init=False, default_factory=lambda: {})
 
+    _collateral_return: Optional[TransactionOutput] = field(init=False, default=None)
+
+    _total_collateral: Optional[int] = field(init=False, default=None)
+
     _inputs_to_redeemers: Dict[UTxO, Redeemer] = field(
         init=False, default_factory=lambda: {}
     )
 
-    _minting_script_to_redeemers: List[Tuple[bytes, Redeemer]] = field(
+    _minting_script_to_redeemers: List[Tuple[ScriptType, Optional[Redeemer]]] = field(
         init=False, default_factory=lambda: []
     )
 
-    _inputs_to_scripts: Dict[UTxO, bytes] = field(
+    _inputs_to_scripts: Dict[UTxO, ScriptType] = field(
         init=False, default_factory=lambda: {}
     )
 
-    _should_estimate_execution_units: bool = field(init=False, default=None)
+    _reference_scripts: List[
+        Union[NativeScript, PlutusV1Script, PlutusV2Script]
+    ] = field(init=False, default_factory=lambda: [])
+
+    _should_estimate_execution_units: Optional[bool] = field(init=False, default=None)
 
     def add_input(self, utxo: UTxO) -> TransactionBuilder:
         """Add a specific UTxO to transaction's inputs.
@@ -166,7 +187,9 @@ class TransactionBuilder:
     def add_script_input(
         self,
         utxo: UTxO,
-        script: bytes,
+        script: Optional[
+            Union[UTxO, NativeScript, PlutusV1Script, PlutusV2Script]
+        ] = None,
         datum: Optional[Datum] = None,
         redeemer: Optional[Redeemer] = None,
     ) -> TransactionBuilder:
@@ -174,7 +197,9 @@ class TransactionBuilder:
 
         Args:
             utxo (UTxO): Script UTxO to be added.
-            script (Optional[bytes]): A plutus script.
+            script (Optional[Union[UTxO, NativeScript, PlutusV1Script, PlutusV2Script]]): A plutus script.
+                If not provided, the script will be inferred from the input UTxO (first arg of this method).
+                The script can also be a specific UTxO whose output contains an inline script.
             datum (Optional[Datum]): A plutus datum to unlock the UTxO.
             redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.
 
@@ -186,42 +211,82 @@ class TransactionBuilder:
                 f"Expect the output address of utxo to be script type, "
                 f"but got {utxo.output.address.address_type} instead."
             )
-        if utxo.output.datum_hash and utxo.output.datum_hash != datum_hash(datum):
+
+        if (
+            utxo.output.datum_hash
+            and datum is not None
+            and utxo.output.datum_hash != datum_hash(datum)
+        ):
             raise InvalidArgumentException(
                 f"Datum hash in transaction output is {utxo.output.datum_hash}, "
                 f"but actual datum hash from input datum is {datum_hash(datum)}."
             )
-        if datum:
+
+        if datum is not None:
             self.datums[datum_hash(datum)] = datum
+
         if redeemer:
+            if redeemer.tag is not None and redeemer.tag != RedeemerTag.SPEND:
+                raise InvalidArgumentException(
+                    f"Expect the redeemer tag's type to be {RedeemerTag.SPEND}, "
+                    f"but got {redeemer.tag} instead."
+                )
+            redeemer.tag = RedeemerTag.SPEND
             self._consolidate_redeemer(redeemer)
             self._inputs_to_redeemers[utxo] = redeemer
-        self._inputs_to_scripts[utxo] = script
+
+        if utxo.output.script:
+            self._inputs_to_scripts[utxo] = utxo.output.script
+            self.reference_inputs.add(utxo)
+            self._reference_scripts.append(utxo.output.script)
+        elif not script:
+            for i in self.context.utxos(str(utxo.output.address)):
+                if i.output.script:
+                    self._inputs_to_scripts[utxo] = i.output.script
+                    self.reference_inputs.add(i)
+                    self._reference_scripts.append(i.output.script)
+                    break
+        elif isinstance(script, UTxO):
+            assert script.output.script is not None
+            self._inputs_to_scripts[utxo] = script.output.script
+            self.reference_inputs.add(script)
+            self._reference_scripts.append(script.output.script)
+        else:
+            self._inputs_to_scripts[utxo] = script
+
         self.inputs.append(utxo)
         return self
 
     def add_minting_script(
         self,
-        script: bytes,
+        script: Union[UTxO, NativeScript, PlutusV1Script, PlutusV2Script],
         redeemer: Optional[Redeemer] = None,
     ) -> TransactionBuilder:
         """Add a minting script along with its datum and redeemer to this transaction.
 
         Args:
-            script (Optional[bytes]): A plutus script.
+            script (Union[UTxO, PlutusV1Script, PlutusV2Script]): A plutus script.
             redeemer (Optional[Redeemer]): A plutus redeemer to unlock the UTxO.
 
         Returns:
             TransactionBuilder: Current transaction builder.
         """
         if redeemer:
-            if redeemer.tag != RedeemerTag.MINT:
+            if redeemer.tag is not None and redeemer.tag != RedeemerTag.MINT:
                 raise InvalidArgumentException(
                     f"Expect the redeemer tag's type to be {RedeemerTag.MINT}, "
                     f"but got {redeemer.tag} instead."
                 )
+            redeemer.tag = RedeemerTag.MINT
             self._consolidate_redeemer(redeemer)
-        self._minting_script_to_redeemers.append((script, redeemer))
+
+        if isinstance(script, UTxO):
+            assert script.output.script is not None
+            self._minting_script_to_redeemers.append((script.output.script, redeemer))
+            self.reference_inputs.add(script)
+            self._reference_scripts.append(script.output.script)
+        else:
+            self._minting_script_to_redeemers.append((script, redeemer))
         return self
 
     def add_input_address(self, address: Union[Address, str]) -> TransactionBuilder:
@@ -255,10 +320,10 @@ class TransactionBuilder:
         Returns:
             TransactionBuilder: Current transaction builder.
         """
-        if datum:
+        if datum is not None:
             tx_out.datum_hash = datum_hash(datum)
         self.outputs.append(tx_out)
-        if add_datum_to_witness:
+        if datum is not None and add_datum_to_witness:
             self.datums[datum_hash(datum)] = datum
         return self
 
@@ -291,11 +356,34 @@ class TransactionBuilder:
         self._fee = fee
 
     @property
-    def scripts(self) -> List[bytes]:
-        return list(
-            set(self._inputs_to_scripts.values())
-            | {s for s, _ in self._minting_script_to_redeemers}
-        )
+    def all_scripts(self) -> List[ScriptType]:
+        scripts: Dict[ScriptHash, ScriptType] = {}
+        s: ScriptType
+
+        if self.native_scripts:
+            for s in self.native_scripts:
+                scripts[script_hash(s)] = s
+
+        for s in self._inputs_to_scripts.values():
+            scripts[script_hash(s)] = s
+
+        for s, _ in self._minting_script_to_redeemers:
+            scripts[script_hash(s)] = s
+
+        return list(scripts.values())
+
+    @property
+    def scripts(self) -> List[ScriptType]:
+        scripts: Dict[ScriptHash, ScriptType] = {
+            script_hash(s): s for s in self.all_scripts
+        }
+        s: ScriptType
+
+        for s in self._reference_scripts:
+            if script_hash(s) in scripts:
+                scripts.pop(script_hash(s))
+
+        return list(scripts.values())
 
     @property
     def datums(self) -> Dict[DatumHash, Datum]:
@@ -303,19 +391,33 @@ class TransactionBuilder:
 
     @property
     def redeemers(self) -> List[Redeemer]:
-        return list(self._inputs_to_redeemers.values()) + [
-            r for _, r in self._minting_script_to_redeemers
+        return [r for r in self._inputs_to_redeemers.values() if r is not None] + [
+            r for _, r in self._minting_script_to_redeemers if r is not None
         ]
 
     @property
     def script_data_hash(self) -> Optional[ScriptDataHash]:
         if self.datums or self.redeemers:
-            return script_data_hash(self.redeemers, list(self.datums.values()))
+            cost_models = {}
+            for s in self.all_scripts:
+                if isinstance(s, PlutusV1Script) or type(s) == bytes:
+                    cost_models[0] = (
+                        self.context.protocol_param.cost_models.get("PlutusV1")
+                        or PLUTUS_V1_COST_MODEL
+                    )
+                if isinstance(s, PlutusV2Script):
+                    cost_models[1] = (
+                        self.context.protocol_param.cost_models.get("PlutusV2")
+                        or PLUTUS_V2_COST_MODEL
+                    )
+            return script_data_hash(
+                self.redeemers, list(self.datums.values()), CostModels(cost_models)
+            )
         else:
             return None
 
     def _calc_change(
-        self, fees, inputs, outputs, address, precise_fee=False
+        self, fees, inputs, outputs, address, precise_fee=False, respect_min_utxo=True
     ) -> List[TransactionOutput]:
         requested = Value(fees)
         for o in outputs:
@@ -350,11 +452,14 @@ class TransactionBuilder:
 
         # when there is only ADA left, simply use remaining coin value as change
         if not change.multi_asset:
-            if change.coin < min_lovelace(change, self.context):
+            if respect_min_utxo and change.coin < min_lovelace_post_alonzo(
+                TransactionOutput(address, change), self.context
+            ):
                 raise InsufficientUTxOBalanceException(
-                    f"Not enough ADA left for change: {change.coin} but needs {min_lovelace(change, self.context)}"
+                    f"Not enough ADA left for change: {change.coin} but needs "
+                    f"{min_lovelace_post_alonzo(TransactionOutput(address, change), self.context)}"
                 )
-            lovelace_change = change.coin
+            lovelace_change = Value(change.coin)
             change_output_arr.append(TransactionOutput(address, lovelace_change))
 
         # If there are multi asset in the change
@@ -369,7 +474,9 @@ class TransactionBuilder:
                 # Combine remainder of provided ADA with last MultiAsset for output
                 # There may be rare cases where adding ADA causes size exceeds limit
                 # We will revisit if it becomes an issue
-                if change.coin < min_lovelace(Value(0, multi_asset), self.context):
+                if respect_min_utxo and change.coin < min_lovelace_post_alonzo(
+                    TransactionOutput(address, Value(0, multi_asset)), self.context
+                ):
                     raise InsufficientUTxOBalanceException(
                         "Not enough ADA left to cover non-ADA assets in a change address"
                     )
@@ -379,7 +486,9 @@ class TransactionBuilder:
                     change_value = Value(change.coin, multi_asset)
                 else:
                     change_value = Value(0, multi_asset)
-                    change_value.coin = min_lovelace(change_value, self.context)
+                    change_value.coin = min_lovelace_post_alonzo(
+                        TransactionOutput(address, change_value), self.context
+                    )
 
                 change_output_arr.append(TransactionOutput(address, change_value))
                 change -= change_value
@@ -407,11 +516,8 @@ class TransactionBuilder:
                 self._outputs += changes
 
         if change_address:
-
             if merge_change:
-
                 for idx, output in enumerate(original_outputs):
-
                     # Find any transaction outputs which already contain the change address
                     if change_address == output.address:
                         if change_output_index is None or output.lovelace == 0:
@@ -420,7 +526,12 @@ class TransactionBuilder:
             # Set fee to max
             self.fee = self._estimate_fee()
             changes = self._calc_change(
-                self.fee, self.inputs, self.outputs, change_address, precise_fee=True
+                self.fee,
+                self.inputs,
+                self.outputs,
+                change_address,
+                precise_fee=True,
+                respect_min_utxo=not merge_change,
             )
 
             _merge_changes(changes)
@@ -431,7 +542,12 @@ class TransactionBuilder:
         if change_address:
             self._outputs = original_outputs
             changes = self._calc_change(
-                self.fee, self.inputs, self.outputs, change_address, precise_fee=True
+                self.fee,
+                self.inputs,
+                self.outputs,
+                change_address,
+                precise_fee=True,
+                respect_min_utxo=not merge_change,
             )
 
             _merge_changes(changes)
@@ -469,7 +585,9 @@ class TransactionBuilder:
         attempt_amount = new_amount + current_amount
 
         # Calculate minimum ada requirements for more precise value size
-        required_lovelace = min_lovelace(attempt_amount, self.context)
+        required_lovelace = min_lovelace_post_alonzo(
+            TransactionOutput(output.address, attempt_amount), self.context
+        )
         attempt_amount.coin = required_lovelace
 
         return len(attempt_amount.to_cbor("bytes")) > max_val_size
@@ -482,10 +600,11 @@ class TransactionBuilder:
     ) -> List[MultiAsset]:
         multi_asset_arr = []
         base_coin = Value(coin=change_estimator.coin)
+        change_address = change_address or Address(FAKE_VKEY.hash())
         output = TransactionOutput(change_address, base_coin)
 
         # iteratively add tokens to output
-        for (policy_id, assets) in change_estimator.multi_asset.items():
+        for policy_id, assets in change_estimator.multi_asset.items():
             temp_multi_asset = MultiAsset()
             temp_value = Value(coin=0)
             temp_assets = Asset()
@@ -526,7 +645,9 @@ class TransactionBuilder:
 
             # Calculate min lovelace required for more precise size
             updated_amount = deepcopy(output.amount)
-            required_lovelace = min_lovelace(updated_amount, self.context)
+            required_lovelace = min_lovelace_post_alonzo(
+                TransactionOutput(change_address, updated_amount), self.context
+            )
             updated_amount.coin = required_lovelace
 
             if len(updated_amount.to_cbor("bytes")) > max_val_size:
@@ -538,6 +659,9 @@ class TransactionBuilder:
         # items at the beginning of next policy to previous policy MultiAssets
         return multi_asset_arr
 
+    def _required_signer_vkey_hashes(self) -> Set[VerificationKeyHash]:
+        return set(self.required_signers) if self.required_signers else set()
+
     def _input_vkey_hashes(self) -> Set[VerificationKeyHash]:
         results = set()
         for i in self.inputs + self.collaterals:
@@ -546,7 +670,6 @@ class TransactionBuilder:
         return results
 
     def _certificate_vkey_hashes(self) -> Set[VerificationKeyHash]:
-
         results = set()
 
         def _check_and_add_vkey(stake_credential: StakeCredential):
@@ -570,7 +693,6 @@ class TransactionBuilder:
         return self.context.protocol_param.key_deposit * len(results)
 
     def _withdrawal_vkey_hashes(self) -> Set[VerificationKeyHash]:
-
         results = set()
 
         if self.withdrawals:
@@ -622,11 +744,11 @@ class TransactionBuilder:
             ):
                 redeemer = self._inputs_to_redeemers[utxo]
                 redeemer.index = sorted_mint_policies.index(
-                    plutus_script_hash(self._inputs_to_scripts[utxo])
+                    script_hash(self._inputs_to_scripts[utxo])
                 )
 
         for script, redeemer in self._minting_script_to_redeemers:
-            redeemer.index = sorted_mint_policies.index(plutus_script_hash(script))
+            redeemer.index = sorted_mint_policies.index(script_hash(script))
 
         self.redeemers.sort(key=lambda r: r.index)
 
@@ -648,11 +770,19 @@ class TransactionBuilder:
             else None,
             certificates=self.certificates,
             withdraws=self.withdrawals,
+            collateral_return=self._collateral_return,
+            total_collateral=self._total_collateral,
+            reference_inputs=[
+                i.input if isinstance(i, UTxO) else i for i in self.reference_inputs
+            ]
+            if self.reference_inputs
+            else None,
         )
         return tx_body
 
     def _build_fake_vkey_witnesses(self) -> List[VerificationKeyWitness]:
         vkey_hashes = self._input_vkey_hashes()
+        vkey_hashes.update(self._required_signer_vkey_hashes())
         vkey_hashes.update(self._native_scripts_vkey_hashes())
         vkey_hashes.update(self._certificate_vkey_hashes())
         vkey_hashes.update(self._withdrawal_vkey_hashes())
@@ -691,9 +821,29 @@ class TransactionBuilder:
         Returns:
             TransactionWitnessSet: A transaction witness set without verification key witnesses.
         """
+
+        native_scripts: List[NativeScript] = []
+        plutus_v1_scripts: List[PlutusV1Script] = []
+        plutus_v2_scripts: List[PlutusV2Script] = []
+
+        for script in self.scripts:
+            if isinstance(script, NativeScript):
+                native_scripts.append(script)
+            elif isinstance(script, PlutusV1Script):
+                plutus_v1_scripts.append(script)
+            elif type(script) is bytes:
+                plutus_v1_scripts.append(PlutusV1Script(script))
+            elif isinstance(script, PlutusV2Script):
+                plutus_v2_scripts.append(script)
+            else:
+                raise InvalidArgumentException(
+                    f"Unsupported script type: {type(script)}"
+                )
+
         return TransactionWitnessSet(
-            native_scripts=self.native_scripts,
-            plutus_script=self.scripts if self.scripts else None,
+            native_scripts=native_scripts if native_scripts else None,
+            plutus_v1_script=plutus_v1_scripts if plutus_v1_scripts else None,
+            plutus_v2_script=plutus_v2_scripts if plutus_v2_scripts else None,
             redeemer=self.redeemers if self.redeemers else None,
             plutus_data=list(self.datums.values()) if self.datums else None,
         )
@@ -724,6 +874,10 @@ class TransactionBuilder:
         self,
         change_address: Optional[Address] = None,
         merge_change: Optional[bool] = False,
+        collateral_change_address: Optional[Address] = None,
+        auto_validity_start_offset: Optional[int] = None,
+        auto_ttl_offset: Optional[int] = None,
+        auto_required_signers: Optional[bool] = None,
     ) -> TransactionBody:
         """Build a transaction body from all constraints set through the builder.
 
@@ -732,12 +886,42 @@ class TransactionBuilder:
                 transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
             merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
                 will be directly added to that transaction output, instead of being added as a separate output.
+            collateral_change_address (Optional[Address]): Address to which collateral changes will be returned.
+            auto_validity_start_offset (Optional[int]): Automatically set the validity start interval of the transaction
+                to the current slot number + the given offset (default -1000).
+                A manually set validity start will always take precedence.
+            auto_ttl_offset (Optional[int]): Automatically set the validity end interval (ttl) of the transaction
+                to the current slot number + the given offset (default 10_000).
+                A manually set ttl will always take precedence.
+            auto_required_signers (Optional[bool]): Automatically add all pubkeyhashes of transaction inputs
+                to required signatories (default only for Smart Contract transactions).
+                Manually set required signers will always take precedence.
 
         Returns:
             TransactionBody: A transaction body.
         """
-        self._update_execution_units()
         self._ensure_no_input_exclusion_conflict()
+
+        # only automatically set the validity interval and required signers if scripts are involved
+        is_smart = bool(self.scripts)
+
+        # Automatically set the validity range to a tight value around transaction creation
+        if (
+            is_smart or auto_validity_start_offset is not None
+        ) and self.validity_start is None:
+            last_slot = self.context.last_block_slot
+            # If None is provided, the default value is -1000
+            if auto_validity_start_offset is None:
+                auto_validity_start_offset = -1000
+            self.validity_start = max(0, last_slot + auto_validity_start_offset)
+
+        if (is_smart or auto_ttl_offset is not None) and self.ttl is None:
+            last_slot = self.context.last_block_slot
+            # If None is provided, the default value is 10_000
+            if auto_ttl_offset is None:
+                auto_ttl_offset = 10_000
+            self.ttl = max(0, last_slot + auto_ttl_offset)
+
         selected_utxos = []
         selected_amount = Value()
         for i in self.inputs:
@@ -785,8 +969,11 @@ class TransactionBuilder:
                 unfulfilled_amount.coin = max(
                     0,
                     unfulfilled_amount.coin
-                    + min_lovelace(
-                        selected_amount - trimmed_selected_amount, self.context
+                    + min_lovelace_post_alonzo(
+                        TransactionOutput(
+                            change_address, selected_amount - trimmed_selected_amount
+                        ),
+                        self.context,
                     ),
                 )
         else:
@@ -807,15 +994,21 @@ class TransactionBuilder:
                         utxo not in selected_utxos
                         and utxo not in self.excluded_inputs
                         and not utxo.output.datum_hash  # UTxO with datum should be added by using `add_script_input`
+                        and not utxo.output.datum
+                        and not utxo.output.script
                     ):
                         additional_utxo_pool.append(utxo)
                         additional_amount += utxo.output.amount
 
-            for i, selector in enumerate(self.utxo_selectors):
+            for index, selector in enumerate(self.utxo_selectors):
                 try:
                     selected, _ = selector.select(
                         additional_utxo_pool,
-                        [TransactionOutput(None, unfulfilled_amount)],
+                        [
+                            TransactionOutput(
+                                Address(FAKE_VKEY.hash()), unfulfilled_amount
+                            )
+                        ],
                         self.context,
                         include_max_fee=False,
                         respect_min_utxo=not can_merge_change,
@@ -827,7 +1020,7 @@ class TransactionBuilder:
                     break
 
                 except UTxOSelectionException as e:
-                    if i < len(self.utxo_selectors) - 1:
+                    if index < len(self.utxo_selectors) - 1:
                         logger.info(e)
                         logger.info(f"{selector} failed. Trying next selector.")
                     else:
@@ -860,20 +1053,127 @@ class TransactionBuilder:
 
         self.inputs[:] = selected_utxos[:]
 
+        # Automatically set the required signers for smart transactions
+        if (
+            is_smart and auto_required_signers is not False
+        ) and self.required_signers is None:
+            # Collect all signatories from explicitly defined
+            # transaction inputs and collateral inputs, and input addresses
+            self.required_signers = list(self._input_vkey_hashes())
+
         self._set_redeemer_index()
 
+        self._set_collateral_return(collateral_change_address or change_address)
+
+        self._update_execution_units(
+            change_address, merge_change, collateral_change_address
+        )
+
         self._add_change_and_fee(change_address, merge_change=merge_change)
 
         tx_body = self._build_tx_body()
 
         return tx_body
 
-    def _update_execution_units(self):
+    def _set_collateral_return(self, collateral_return_address: Optional[Address]):
+        """Calculate and set the change returned from the collateral inputs.
+
+        Args:
+            collateral_return_address (Address): Address to which the collateral change will be returned.
+        """
+        witnesses = self._build_fake_witness_set()
+
+        # Make sure there is at least one script input
+        if (
+            not witnesses.plutus_v1_script
+            and not witnesses.plutus_v2_script
+            and not self._reference_scripts
+        ):
+            return
+
+        if not collateral_return_address:
+            return
+
+        collateral_amount = (
+            max_tx_fee(context=self.context)
+            * self.context.protocol_param.collateral_percent
+            // 100
+        )
+
+        if not self.collaterals:
+            tmp_val = Value()
+
+            def _add_collateral_input(cur_total, candidate_inputs):
+                while cur_total.coin < collateral_amount and candidate_inputs:
+                    candidate = candidate_inputs.pop()
+                    if (
+                        not candidate.output.address.address_type.name.startswith(
+                            "SCRIPT"
+                        )
+                        and candidate.output.amount.coin > 2000000
+                    ):
+                        self.collaterals.append(candidate)
+                        cur_total += candidate.output.amount
+
+            sorted_inputs = sorted(
+                self.inputs.copy(),
+                key=lambda i: (len(i.output.to_cbor()), -i.output.amount.coin),
+            )
+            _add_collateral_input(tmp_val, sorted_inputs)
+
+            if tmp_val.coin < collateral_amount:
+                sorted_inputs = sorted(
+                    self.context.utxos(str(collateral_return_address)),
+                    key=lambda i: (len(i.output.to_cbor()), -i.output.amount.coin),
+                )
+                _add_collateral_input(tmp_val, sorted_inputs)
+
+        total_input = Value()
+
+        for utxo in self.collaterals:
+            total_input += utxo.output.amount
+
+        if collateral_amount > total_input.coin:
+            raise ValueError(
+                f"Minimum collateral amount {collateral_amount} is greater than total "
+                f"provided collateral inputs {total_input}"
+            )
+        else:
+            return_amount = total_input - collateral_amount
+            min_lovelace_val = min_lovelace_post_alonzo(
+                TransactionOutput(collateral_return_address, return_amount),
+                self.context,
+            )
+            if min_lovelace_val > return_amount.coin:
+                raise ValueError(
+                    f"Minimum lovelace amount for collateral return {min_lovelace_val} is "
+                    f"greater than collateral change {return_amount.coin}. Please provide more collateral inputs."
+                )
+            else:
+                self._collateral_return = TransactionOutput(
+                    collateral_return_address, total_input - collateral_amount
+                )
+                self._total_collateral = collateral_amount
+
+    def _update_execution_units(
+        self,
+        change_address: Optional[Address] = None,
+        merge_change: Optional[bool] = False,
+        collateral_change_address: Optional[Address] = None,
+    ):
         if self._should_estimate_execution_units:
-            estimated_execution_units = self._estimate_execution_units()
+            estimated_execution_units = self._estimate_execution_units(
+                change_address, merge_change, collateral_change_address
+            )
             for r in self.redeemers:
+                assert (
+                    r.tag is not None
+                ), "Expected tag of redeemer to be set, but found None"
                 key = f"{r.tag.name.lower()}:{r.index}"
-                if key not in estimated_execution_units:
+                if (
+                    key not in estimated_execution_units
+                    or estimated_execution_units[key] is None
+                ):
                     raise TransactionBuilderException(
                         f"Cannot find execution unit for redeemer: {r} "
                         f"in estimated execution units: {estimated_execution_units}"
@@ -889,15 +1189,19 @@ class TransactionBuilder:
     def _estimate_execution_units(
         self,
         change_address: Optional[Address] = None,
-    ):
-        # Create a shallow copy of current builder, so we won't mess up current builder's internal states
+        merge_change: Optional[bool] = False,
+        collateral_change_address: Optional[Address] = None,
+    ) -> Dict[str, ExecutionUnits]:
+        # Create a deep copy of current builder, so we won't mess up current builder's internal states
         tmp_builder = TransactionBuilder(self.context)
         for f in fields(self):
             if f.name not in ("context",):
-                setattr(tmp_builder, f.name, copy(getattr(self, f.name)))
+                setattr(tmp_builder, f.name, deepcopy(getattr(self, f.name)))
         tmp_builder._should_estimate_execution_units = False
         self._should_estimate_execution_units = False
-        tx_body = tmp_builder.build(change_address)
+        tx_body = tmp_builder.build(
+            change_address, merge_change, collateral_change_address
+        )
         witness_set = tmp_builder._build_fake_witness_set()
         tx = Transaction(
             tx_body, witness_set, auxiliary_data=tmp_builder.auxiliary_data
@@ -910,6 +1214,10 @@ class TransactionBuilder:
         signing_keys: List[Union[SigningKey, ExtendedSigningKey]],
         change_address: Optional[Address] = None,
         merge_change: Optional[bool] = False,
+        collateral_change_address: Optional[Address] = None,
+        auto_validity_start_offset: Optional[int] = None,
+        auto_ttl_offset: Optional[int] = None,
+        auto_required_signers: Optional[bool] = None,
     ) -> Transaction:
         """Build a transaction body from all constraints set through the builder and sign the transaction with
         provided signing keys.
@@ -921,16 +1229,33 @@ class TransactionBuilder:
                 transaction body will likely be unbalanced (sum of inputs is greater than the sum of outputs).
             merge_change (Optional[bool]): If the change address match one of the transaction output, the change amount
                 will be directly added to that transaction output, instead of being added as a separate output.
+            collateral_change_address (Optional[Address]): Address to which collateral changes will be returned.
+            auto_validity_start_offset (Optional[int]): Automatically set the validity start interval of the transaction
+                to the current slot number + the given offset (default -1000).
+                A manually set validity start will always take precedence.
+            auto_ttl_offset (Optional[int]): Automatically set the validity end interval (ttl) of the transaction
+                to the current slot number + the given offset (default 10_000).
+                A manually set ttl will always take precedence.
+            auto_required_signers (Optional[bool]): Automatically add all pubkeyhashes of transaction inputs
+                to required signatories (default only for Smart Contract transactions).
+                Manually set required signers will always take precedence.
 
         Returns:
             Transaction: A signed transaction.
         """
 
-        tx_body = self.build(change_address=change_address, merge_change=merge_change)
+        tx_body = self.build(
+            change_address=change_address,
+            merge_change=merge_change,
+            collateral_change_address=collateral_change_address,
+            auto_validity_start_offset=auto_validity_start_offset,
+            auto_ttl_offset=auto_ttl_offset,
+            auto_required_signers=auto_required_signers,
+        )
         witness_set = self.build_witness_set()
         witness_set.vkey_witnesses = []
 
-        for signing_key in signing_keys:
+        for signing_key in set(signing_keys):
             signature = signing_key.sign(tx_body.hash())
             witness_set.vkey_witnesses.append(
                 VerificationKeyWitness(signing_key.to_verification_key(), signature)
